@@ -1,34 +1,25 @@
 //
 //  Little cms
-//  Copyright (C) 1998-2002 Marti Maria
+//  Copyright (C) 1998-2003 Marti Maria
 //
-// THIS SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
-// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
-// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
+// Permission is hereby granted, free of charge, to any person obtaining 
+// a copy of this software and associated documentation files (the "Software"), 
+// to deal in the Software without restriction, including without limitation 
+// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
+// and/or sell copies of the Software, and to permit persons to whom the Software 
+// is furnished to do so, subject to the following conditions:
 //
-// IN NO EVENT SHALL MARTI MARIA BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
-// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
-// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
-// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
-// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
-// OF THIS SOFTWARE.
+// The above copyright notice and this permission notice shall be included in 
+// all copies or substantial portions of the Software.
 //
-//
-// This library is free software; you can redistribute it and/or
-// modify it under the terms of the GNU Lesser General Public
-// License as published by the Free Software Foundation; either
-// version 2 of the License, or (at your option) any later version.
-//
-// This library is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-// Lesser General Public License for more details.
-//
-// You should have received a copy of the GNU Lesser General Public
-// License along with this library; if not, write to the Free Software
-// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
+// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
+// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
+// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
+// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
+// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION 
+// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-// Version 1.10c
 
 // This program does apply profiles to (some) JPEG files
 
@@ -127,6 +118,13 @@ void FatalError(const char *frm, ...)
        exit(1);
 }
 
+static
+int MyErrorHandler(int ErrorCode, const char *ErrorText)
+{
+    FatalError("%s", ErrorText);
+    return 0;
+}
+
 
 // Out of mem
 static
@@ -364,6 +362,7 @@ int GetProfileColorSpace(cmsHPROFILE hProfile)
 
        case icSigHexachromeData: return PT_HiFi;
 
+
        default:  return icMaxEnumData;
        }
 }
@@ -432,11 +431,11 @@ void WriteOutputFields(int OutputColorSpace)
         return;
     }
     
-    Compressor.in_color_space = n;
-    Compressor.jpeg_color_space = m;
+    Compressor.in_color_space = (J_COLOR_SPACE) n;
+    Compressor.jpeg_color_space = (J_COLOR_SPACE) m;
     Compressor.input_components = Compressor.num_components = k;
     jpeg_set_defaults(&Compressor);
-    jpeg_set_colorspace(&Compressor, m);
+    jpeg_set_colorspace(&Compressor, (J_COLOR_SPACE) m);
     jpeg_set_quality(&Compressor, jpegQuality,1);    
 }
 
@@ -447,14 +446,23 @@ int DoTransform(cmsHTRANSFORM hXForm)
     JSAMPROW ScanLineOut;
 
 
-      
+    
+       //Preserve resolution values from the original
+       // (Thanks to robert bergs for finding out this bug)
+
+       Compressor.density_unit=Decompressor.density_unit;
+       Compressor.X_density=Decompressor.X_density;
+       Compressor.Y_density=Decompressor.Y_density;
+
+       Compressor.write_JFIF_header = 1;
+     
        jpeg_start_decompress(&Decompressor);
        jpeg_start_compress(&Compressor, TRUE);
 
       // FIXME: # of bytes is *4 at most, but can be less
 
-       ScanLineIn  = malloc(Decompressor.output_width * 4);
-       ScanLineOut = malloc(Compressor.image_width * 4);
+       ScanLineIn  = (JSAMPROW) malloc(Decompressor.output_width * 4);
+       ScanLineOut = (JSAMPROW) malloc(Compressor.image_width * 4);
 
        while (Decompressor.output_scanline <
                             Decompressor.output_height) {
@@ -498,9 +506,7 @@ int TransformImage(char *cDefInpProf, char *cOutProf)
 
        if (BlackWhiteCompensation) {
 
-            dwFlags |= cmsFLAGS_WHITEBLACKCOMPENSATION;
-            if (PrecalcMode == 0) 
-                    FatalError("Cannot use White/Black compensation without precalculation. Use /c1 or /c2.");
+            dwFlags |= cmsFLAGS_BLACKPOINTCOMPENSATION;            
        }
 
        switch (PrecalcMode) {
@@ -581,7 +587,7 @@ int TransformImage(char *cDefInpProf, char *cOutProf)
                                           hProof, Intent, 
                                           ProofingIntent, dwFlags);
 
-       // Handle tile by tile or strip by strip
+       // Handle tile by tile or strip by strip strtok
 
        DoTransform(xform);
 
@@ -602,7 +608,7 @@ int TransformImage(char *cDefInpProf, char *cOutProf)
 static
 void Help(int level)
 {
-     fprintf(stderr, "little cms ICC profile applier for JPEG - v1.4\n\n");
+     fprintf(stderr, "little cms ICC profile applier for JPEG - v1.5\n\n");
 
      switch(level) {
 
@@ -619,7 +625,7 @@ void Help(int level)
      fprintf(stderr, "%cm<0,1,2,3> - SoftProof intent\n", SW);
      fprintf(stderr, "\n");
 
-     fprintf(stderr, "%cb - Black/White compensation\n", SW);
+     fprintf(stderr, "%cb - Black point compensation\n", SW);
      fprintf(stderr, "%cn - Ignore embedded profile\n", SW);
 
 
@@ -786,6 +792,8 @@ int main(int argc, char* argv[])
 
       HandleSwitches(argc, argv);
 
+      cmsSetErrorHandler(MyErrorHandler);
+
       if ((argc - xoptind) != 2) {
 
               Help(0);              
