@@ -1,32 +1,25 @@
 //
 //  Little cms
-//  Copyright (C) 1998-2002 Marti Maria
+//  Copyright (C) 1998-2004 Marti Maria
 //
-// THIS SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
-// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
-// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
+// Permission is hereby granted, free of charge, to any person obtaining 
+// a copy of this software and associated documentation files (the "Software"), 
+// to deal in the Software without restriction, including without limitation 
+// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
+// and/or sell copies of the Software, and to permit persons to whom the Software 
+// is furnished to do so, subject to the following conditions:
 //
-// IN NO EVENT SHALL MARTI MARIA BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
-// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
-// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
-// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
-// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
-// OF THIS SOFTWARE.
+// The above copyright notice and this permission notice shall be included in 
+// all copies or substantial portions of the Software.
 //
-//
-// This library is free software; you can redistribute it and/or
-// modify it under the terms of the GNU Lesser General Public
-// License as published by the Free Software Foundation; either
-// version 2 of the License, or (at your option) any later version.
-//
-// This library is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-// Lesser General Public License for more details.
-//
-// You should have received a copy of the GNU Lesser General Public
-// License along with this library; if not, write to the Free Software
-// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
+// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
+// THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
+// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
+// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
+// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION 
+// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 
 #include "lcms.h"
 
@@ -309,8 +302,8 @@ void PrecalculatedXFORM(_LPcmsTRANSFORM p,
        output = (LPBYTE) out;
        n = Size;                    // Buffer len
 
-       for (i=0; i < n; i++)
-       {
+       for (i=0; i < n; i++) {
+
        accum = p -> FromInput(p, wIn, accum);
        cmsEvalLUT(p -> DeviceLink, wIn, wOut);
        output = p -> ToOutput(p, wOut, output);
@@ -318,6 +311,49 @@ void PrecalculatedXFORM(_LPcmsTRANSFORM p,
 }
 
 
+// Using precalculated LUT + Cache
+
+static
+void CachedXFORM(_LPcmsTRANSFORM p,
+                     LPVOID in,
+                     LPVOID out, unsigned int Size)
+{
+       register LPBYTE accum;
+       register LPBYTE output;
+       WORD wIn[MAXCHANNELS], wOut[MAXCHANNELS];
+       register unsigned int i, n;
+
+
+       accum  = (LPBYTE) in;
+       output = (LPBYTE) out;
+       n = Size;                    // Buffer len
+
+       // Empty buffers for quick memcmp
+
+       ZeroMemory(wIn,  sizeof(WORD) * MAXCHANNELS);
+       ZeroMemory(wOut, sizeof(WORD) * MAXCHANNELS);
+
+       for (i=0; i < n; i++) {
+
+       accum = p -> FromInput(p, wIn, accum);
+       
+       if (memcmp(wIn, p ->CacheIn, sizeof(WORD) * MAXCHANNELS) == 0) {
+
+            CopyMemory(wOut, p ->CacheOut, sizeof(WORD) * MAXCHANNELS);
+       }
+       else {
+
+            CopyMemory(p ->CacheIn, wIn, sizeof(WORD) * MAXCHANNELS);
+
+            cmsEvalLUT(p -> DeviceLink, wIn, wOut);
+            
+            CopyMemory(p ->CacheOut, wOut, sizeof(WORD) * MAXCHANNELS);
+       }
+
+       output = p -> ToOutput(p, wOut, output);
+       }
+}
+
 // Using smelted Matrix/Shaper
 
 static
@@ -574,36 +610,18 @@ BOOL cmsBuildSmeltMatShaper(_LPcmsTRANSFORM p)
 
 // Conversion between PCS ------------------------------------------
 
-// Identifies intent archieved by LUT, only absolute colorimetric ones
-// are handled separatly
+// Identifies intent archieved by LUT
 
 static
 int GetPhase(cmsHPROFILE hProfile)
-{
-       icColorSpaceSignature PCS;
-       icRenderingIntent Intent;
+{       
+       switch (cmsGetPCS(hProfile)) {
 
-       PCS    = cmsGetPCS(hProfile);
-       Intent = (icRenderingIntent) cmsTakeRenderingIntent(hProfile);
+       case icSigXYZData: return XYZRel;
 
-       switch (PCS) {
-       case icSigXYZData:
-                     if (Intent == icAbsoluteColorimetric)
+       case icSigLabData: return LabRel;
 
-                            return XYZAbs;
-                     else
-                            return XYZRel;
-
-       case icSigLabData:
-
-                     if (Intent == icAbsoluteColorimetric)
-
-                            return LabAbs;
-                     else
-                            return LabRel;
-
-       default:
-                     cmsSignalError(LCMS_ERRC_ABORTED, "Invalid PCS");
+       default:  cmsSignalError(LCMS_ERRC_ABORTED, "Invalid PCS");
        }
 
        return XYZRel;
@@ -632,16 +650,57 @@ void TakeConversionRoutines(_LPcmsTRANSFORM p, int DoBPC)
        cmsTakeMediaBlackPoint(&BlackPointOut,  p -> OutputProfile);
        cmsReadChromaticAdaptationMatrix(&ChromaticAdaptationMatrixOut, p -> OutputProfile);
 
-       // Guess black point from intent table
-       if (DoBPC) {
-       
-           cmsDetectBlackPoint(&BlackPointIn,    p->InputProfile,   p->Intent);
-           cmsDetectBlackPoint(&BlackPointOut,   p->OutputProfile,  p->Intent);          
-       }
-      
+            
        if (p -> Preview == NULL)     // Non-proofing
-       {
-              cmsChooseCnvrt(p -> Intent == INTENT_ABSOLUTE_COLORIMETRIC,
+       {            
+            if (p ->Intent == INTENT_PERCEPTUAL ||
+                p ->Intent == INTENT_SATURATION) {
+
+
+                    // For v4 profiles, Perceptual PCS has a reference black point
+                    // which v2 profiles should scale to.
+
+                    if ((cmsGetProfileICCversion(p ->InputProfile) >= 0x4000000) ||
+                        (cmsGetProfileICCversion(p ->OutputProfile) >= 0x4000000)) {
+
+                           DoBPC = TRUE;
+                    }
+            }
+
+            // Black point compensation does not apply to absolute intent
+
+            if (p ->Intent == INTENT_ABSOLUTE_COLORIMETRIC) 
+                            DoBPC = FALSE;
+
+            // Black point compensation does not apply to devicelink or abstract profiles
+
+            if (cmsGetDeviceClass(p ->InputProfile) == icSigAbstractClass ||
+                cmsGetDeviceClass(p ->InputProfile) == icSigLinkClass)
+                            DoBPC = FALSE;
+
+            if (cmsGetDeviceClass(p ->OutputProfile) == icSigAbstractClass ||
+                cmsGetDeviceClass(p ->OutputProfile) == icSigLinkClass)
+                            DoBPC = FALSE;
+
+
+            if (DoBPC) {
+       
+                // Detect Black points
+
+                cmsDetectBlackPoint(&BlackPointIn,    p->InputProfile,   p->Intent, 0);
+                cmsDetectBlackPoint(&BlackPointOut,   p->OutputProfile,  p->Intent, 0);          
+
+                // If equal black points, then do nothing. This often applies to BP=0
+
+                if (BlackPointIn.X == BlackPointOut.X &&
+                    BlackPointIn.Y == BlackPointOut.Y &&
+                    BlackPointIn.Z == BlackPointOut.Z) 
+                                DoBPC = FALSE;
+
+
+            }
+            
+            cmsChooseCnvrt(p -> Intent == INTENT_ABSOLUTE_COLORIMETRIC,
 
                  p -> Phase1,
                              &BlackPointIn,
@@ -667,12 +726,26 @@ void TakeConversionRoutines(_LPcmsTRANSFORM p, int DoBPC)
        cmsTakeIluminant(&IlluminantProof,        p -> PreviewProfile);
        cmsTakeMediaWhitePoint(&WhitePointProof,  p -> PreviewProfile);
        cmsTakeMediaBlackPoint(&BlackPointProof,  p -> PreviewProfile);
-
-       if (DoBPC) 
-            cmsDetectBlackPoint(&BlackPointProof, p->PreviewProfile, p->Intent);
-
        cmsReadChromaticAdaptationMatrix(&ChromaticAdaptationMatrixProof, p -> PreviewProfile);
 
+       if (DoBPC) {
+
+            cmsDetectBlackPoint(&BlackPointProof, p->PreviewProfile, p->Intent, 0);
+            cmsDetectBlackPoint(&BlackPointIn,    p->InputProfile,   p->Intent, 0);
+            cmsDetectBlackPoint(&BlackPointOut,   p->OutputProfile,  p->Intent, 0);  
+            
+            // Reality check
+
+            if (BlackPointIn.X == BlackPointProof.X &&
+                BlackPointIn.Y == BlackPointProof.Y &&
+                BlackPointIn.Z == BlackPointProof.Z) 
+                                DoBPC = FALSE;
+
+
+       }
+
+       
+
        cmsChooseCnvrt(p -> Intent == INTENT_ABSOLUTE_COLORIMETRIC,
 
                  p -> Phase1,
@@ -706,8 +779,6 @@ void TakeConversionRoutines(_LPcmsTRANSFORM p, int DoBPC)
                  0,
                  &p->Stage2,
                  &p->m2, &p->of2);
-
-
        }
 
 }
@@ -716,13 +787,16 @@ void TakeConversionRoutines(_LPcmsTRANSFORM p, int DoBPC)
 // Check colorspace
 
 static
-BOOL IsProperColorSpace(cmsHPROFILE hProfile, DWORD dwFormat)
+BOOL IsProperColorSpace(cmsHPROFILE hProfile, DWORD dwFormat, BOOL lUsePCS)
 {
        int Space = T_COLORSPACE(dwFormat);
 
        if (Space == PT_ANY) return TRUE;
 
-       return (_cmsICCcolorSpace(Space) == cmsGetColorSpace(hProfile));
+       if (lUsePCS)
+           return (_cmsICCcolorSpace(Space) == cmsGetPCS(hProfile));
+       else
+           return (_cmsICCcolorSpace(Space) == cmsGetColorSpace(hProfile));
 }
 
 
@@ -758,20 +832,43 @@ _LPcmsTRANSFORM AllocEmptyTransform(void)
     p -> OutMatShaper   = NULL;
     p -> SmeltMatShaper = NULL;
     p -> NamedColorList = NULL;
-    p -> EntryColorSpace = 0;
-    p -> ExitColorSpace  = 0;
+    p -> EntryColorSpace = (icColorSpaceSignature) 0;
+    p -> ExitColorSpace  = (icColorSpaceSignature) 0;
 
     return p;
 }
 
 
+// Identify whatever a transform is to be cached
+
+static
+void SetPrecalculatedTransform(_LPcmsTRANSFORM p, DWORD dwFlags)
+{
+    p -> xform = PrecalculatedXFORM;
+        
+    if (!(dwFlags & cmsFLAGS_NOTCACHE)) {
+        
+        ZeroMemory(p ->CacheIn, sizeof(WORD) * MAXCHANNELS);
+        cmsEvalLUT(p ->DeviceLink, p->CacheIn, p ->CacheOut);        
+        p ->xform = CachedXFORM;        
+    }
+    
+}
+
+
 // Transform is identified as device-link
 static 
-cmsHPROFILE CreateDeviceLinkTransform(_LPcmsTRANSFORM p)
+cmsHPROFILE CreateDeviceLinkTransform(_LPcmsTRANSFORM p, DWORD dwFlags)
 {
     
-    if (!IsProperColorSpace(p->InputProfile, p->InputFormat)) {
-        cmsSignalError(LCMS_ERRC_WARNING, "Device link is operating on wrong colorspace");
+    if (!IsProperColorSpace(p->InputProfile, p->InputFormat, FALSE)) {
+        cmsSignalError(LCMS_ERRC_ABORTED, "Device link is operating on wrong colorspace on input");
+        return NULL;
+    }
+
+    if (!IsProperColorSpace(p->InputProfile, p->OutputFormat, TRUE)) {
+        cmsSignalError(LCMS_ERRC_ABORTED, "Device link is operating on wrong colorspace on output");
+        return NULL;
     }
 
     // Device link does only have AToB0Tag (ICC-Spec 1998/09)
@@ -796,8 +893,12 @@ cmsHPROFILE CreateDeviceLinkTransform(_LPcmsTRANSFORM p)
     p -> Phase1 = -1;
     p -> Phase2 = -1;
     p -> Phase3 = -1;
-    p -> xform = PrecalculatedXFORM;
+
+    SetPrecalculatedTransform(p, dwFlags);
+    
+
     p -> ExitColorSpace = cmsGetPCS(p -> InputProfile);
+    
 
     // Precalculated device-link profile is ready
     return (cmsHTRANSFORM) p;
@@ -818,9 +919,13 @@ void CreateProof(_LPcmsTRANSFORM p,
       // Apr-15, 2002 - Too much profiles does have bogus content
       // on preview tag, so I do compute it by my own.
 
-      p -> Preview = _cmsComputeSoftProofLUT(p ->PreviewProfile, p ->ProofIntent); 
+      p -> Preview = _cmsComputeSoftProofLUT(p ->PreviewProfile, p ->Intent); 
       p -> Phase2  = LabRel;
 
+      // That's a proofing transfor, so use second intent for output.
+
+      *ToTagPtr  = PCS2Device[p->ProofIntent];
+
       if (p -> Preview == NULL) {
           
         ProofTag = Preview[p -> Intent];
@@ -836,8 +941,7 @@ void CreateProof(_LPcmsTRANSFORM p,
 
              p -> Preview = cmsReadICCLut(p ->PreviewProfile, ProofTag);
              p -> Phase2 = GetPhase(p ->PreviewProfile);
-             *ToTagPtr  = PCS2Device[p->ProofIntent];
-             
+                         
         }
         else
              {
@@ -856,7 +960,7 @@ void CreateProof(_LPcmsTRANSFORM p,
     if (dwFlags & cmsFLAGS_GAMUTCHECK) {
 
                    
-             p -> Gamut = _cmsComputeGamutLUT(p->PreviewProfile, p ->ProofIntent);                               
+             p -> Gamut = _cmsComputeGamutLUT(p->PreviewProfile, p ->Intent);                               
              if (p -> Gamut == NULL) {
 
                 // Profile goes only in one direction... try to see 
@@ -889,29 +993,7 @@ _LPcmsTRANSFORM PickTransformRoutine(_LPcmsTRANSFORM p,
                                     icTagSignature *ToTagPtr)
 {
 
-        // Can we optimize matrix-shaper only transform?
-
-       if (*FromTagPtr == 0 && 
-           *ToTagPtr == 0 && 
-           !p->PreviewProfile  && 
-           p -> Intent != INTENT_ABSOLUTE_COLORIMETRIC && 
-           (p -> EntryColorSpace == icSigRgbData) && 
-           (p -> ExitColorSpace == icSigRgbData)) {
-
-              // Yes... try to smelt matrix-shapers
-              p -> xform = MatrixShaperXFORM;
-              *dwFlagsPtr |= cmsFLAGS_NOTPRECALC;
-
-              if (!cmsBuildSmeltMatShaper(p))
-              {
-                     cmsSignalError(LCMS_ERRC_ABORTED, "unable to smelt shaper-matrix, required tags missing");               
-                     return NULL;
-              }
-
-              p -> Phase1 = p -> Phase3 = XYZRel;
-              return p;
-
-       }
+       
        
 
         // Is a named color profile?
@@ -921,6 +1003,29 @@ _LPcmsTRANSFORM PickTransformRoutine(_LPcmsTRANSFORM p,
                   p ->FromDevice = NC2toPCS;
         }
         else {
+                // Can we optimize matrix-shaper only transform?
+
+                   if (*FromTagPtr == 0 && 
+                       *ToTagPtr == 0 && 
+                       !p->PreviewProfile  && 
+                       p -> Intent != INTENT_ABSOLUTE_COLORIMETRIC && 
+                       (p -> EntryColorSpace == icSigRgbData) && 
+                       (p -> ExitColorSpace == icSigRgbData)) {
+
+                          // Yes... try to smelt matrix-shapers
+                          p -> xform = MatrixShaperXFORM;
+                          *dwFlagsPtr |= cmsFLAGS_NOTPRECALC;
+
+                          if (!cmsBuildSmeltMatShaper(p))
+                          {
+                                 cmsSignalError(LCMS_ERRC_ABORTED, "unable to smelt shaper-matrix, required tags missing");               
+                                 return NULL;
+                          }
+
+                          p -> Phase1 = p -> Phase3 = XYZRel;
+                          return p;
+
+                }
 
                 // No, is a transform involving LUT
 
@@ -997,10 +1102,10 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        icTagSignature ToTag;
        
        if (nIntent < 0 || nIntent > 3 ||
-           ProofingIntent < 0 || ProofingIntent > 3)
-       {
-       cmsSignalError(LCMS_ERRC_ABORTED, "cmsCreateTransform: intent mismatch");
-       return NULL;
+           ProofingIntent < 0 || ProofingIntent > 3) {
+
+            cmsSignalError(LCMS_ERRC_ABORTED, "cmsCreateTransform: intent mismatch");
+            return NULL;
        }
 
        p = AllocEmptyTransform();
@@ -1015,6 +1120,7 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        p -> PreviewProfile = ProofingProfile;            
        p -> InputFormat    = InputFormat;
        p -> OutputFormat   = OutputFormat;
+       p -> lInputV4Lab    = p ->lOutputV4Lab = FALSE;
 
        p -> FromInput = _cmsIdentifyInputFormat(p, InputFormat);
        p -> ToOutput  = _cmsIdentifyOutputFormat(p, OutputFormat);
@@ -1040,13 +1146,14 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        //  Device link are means to store precalculated transform grids.
        if (cmsGetDeviceClass(InputProfile) == icSigLinkClass) {
 
-            return CreateDeviceLinkTransform(p);
+            return CreateDeviceLinkTransform(p, dwFlags);
        }
+         
+       if (!IsProperColorSpace(InputProfile, InputFormat, FALSE)) {
 
-     
-     
-       if (!IsProperColorSpace(InputProfile, InputFormat)) {
-              cmsSignalError(LCMS_ERRC_WARNING, "Input profile is operating on wrong colorspace");
+              cmsSignalError(LCMS_ERRC_ABORTED, "Input profile is operating on wrong colorspace");
+              cmsDeleteTransform((cmsHTRANSFORM) p);
+              return NULL;
        }
 
        p ->EntryColorSpace = cmsGetColorSpace(InputProfile);
@@ -1055,7 +1162,7 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        if (cmsGetDeviceClass(InputProfile) == icSigNamedColorClass) {
 
         if (p ->NamedColorList == NULL)
-                p ->NamedColorList = cmsAllocNamedColorList();
+                p ->NamedColorList = cmsAllocNamedColorList(0);
 
         cmsReadICCnamedColorList(p, InputProfile, icSigNamedColor2Tag);
         
@@ -1083,8 +1190,10 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        }
 
             
-       if (!IsProperColorSpace(OutputProfile, OutputFormat)) {
-              cmsSignalError(LCMS_ERRC_WARNING, "Output profile is operating on wrong colorspace");
+       if (!IsProperColorSpace(OutputProfile, OutputFormat, FALSE)) {
+              cmsSignalError(LCMS_ERRC_ABORTED, "Output profile is operating on wrong colorspace");
+              cmsDeleteTransform((cmsHTRANSFORM) p);             
+              return NULL;
        }
 
        p -> ExitColorSpace = cmsGetColorSpace(OutputProfile);
@@ -1106,8 +1215,8 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        FromTag  = Device2PCS[nIntent];
        ToTag    = PCS2Device[nIntent];
 
-       if (!cmsIsTag(InputProfile, FromTag))
-       {
+       if (!cmsIsTag(InputProfile, FromTag)) {
+
               FromTag = Device2PCS[0];
 
               if (!cmsIsTag(InputProfile,  FromTag)) {
@@ -1120,10 +1229,19 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
            CreateProof(p, dwFlags, &ToTag);                       
        
 
-
        if (!cmsIsTag(OutputProfile,  ToTag)) {
-              ToTag = PCS2Device[0];
-              if (!cmsIsTag(OutputProfile,  ToTag))
+
+           ToTag = PCS2Device[0];
+
+           // 12-Dec-2003, Abstract profiles can be placed as output and still using AToB0          
+           if (cmsGetDeviceClass(OutputProfile) == icSigAbstractClass) {
+           
+                    if (!cmsIsTag(OutputProfile,  ToTag)) {
+                                ToTag = (icTagSignature) icSigAToB0Tag;
+                    }           
+           }
+                
+           if (!cmsIsTag(OutputProfile,  ToTag))
                             ToTag = (icTagSignature)0;
        }
 
@@ -1134,8 +1252,11 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        if (dwFlags & cmsFLAGS_MATRIXOUTPUT)
               ToTag = (icTagSignature)0;
 
-       if (dwFlags & cmsFLAGS_GAMUTCHECK)
+       if (dwFlags & cmsFLAGS_GAMUTCHECK) {
+
               p -> DoGamutCheck = TRUE;
+              dwFlags |= cmsFLAGS_NOTPRECALC;
+       }
 
      
        if (PickTransformRoutine(p, &dwFlags, &FromTag, &ToTag) == NULL) {
@@ -1145,15 +1266,37 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
 
        }
 
-       TakeConversionRoutines(p, dwFlags & cmsFLAGS_WHITEBLACKCOMPENSATION);
+       TakeConversionRoutines(p, dwFlags & cmsFLAGS_BLACKPOINTCOMPENSATION);
 
      
        if (!(dwFlags & cmsFLAGS_NOTPRECALC)) {
 
                LPLUT DeviceLink;    
-
+               
                DeviceLink = _cmsPrecalculateDeviceLink((cmsHTRANSFORM) p, dwFlags);
-                                        
+                
+               // If input colorspace is Rgb, Cmy, then use tetrahedral interpolation
+               // for speed reasons (it only works well on spaces on Luma is diagonal, and 
+               // not if luma is in separate channel)
+
+               if (p ->EntryColorSpace == icSigRgbData ||
+                   p ->EntryColorSpace == icSigCmyData) {
+                   
+
+                    DeviceLink -> CLut16params.Interp3D = cmsTetrahedralInterp16;
+               }
+               
+               // If this is a 8-bit transform, optimize LUT further.
+                                           
+                if ((T_BYTES(InputFormat) == 1) && (T_CHANNELS(InputFormat) == 3)) {
+
+                    DeviceLink = _cmsBlessLUT8(DeviceLink);
+                    if (DeviceLink == NULL) return NULL;
+
+                }
+                
+
+
                 if (DeviceLink) {
 
                     p ->DeviceLink = DeviceLink;
@@ -1170,7 +1313,8 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
                 }
 
 
-              p -> xform = PrecalculatedXFORM;
+              SetPrecalculatedTransform(p, dwFlags);
+              
 
        }
 
@@ -1317,7 +1461,7 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
                                                                 DWORD dwFlags)
 {
     cmsHTRANSFORM Transforms[257];
-    DWORD dwPrecalcFlags = (dwFlags|cmsFLAGS_NOTPRECALC);
+    DWORD dwPrecalcFlags = (dwFlags|cmsFLAGS_NOTPRECALC|cmsFLAGS_NOTCACHE);
     DWORD FormatInput, FormatOutput;
     cmsHPROFILE hLab, hXYZ, hProfile;   
     icColorSpaceSignature ColorSpace, CurrentColorSpace;
@@ -1333,7 +1477,8 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
     }
 
     // Creates a phantom transform for latter filling
-    p = cmsCreateTransform(NULL, dwInput, NULL, dwOutput, Intent, cmsFLAGS_NULLTRANSFORM);
+    p = (_LPcmsTRANSFORM) cmsCreateTransform(NULL, dwInput, 
+                                             NULL, dwOutput, Intent, cmsFLAGS_NULLTRANSFORM);
 
     // If user wants null one, give it
     if (dwFlags & cmsFLAGS_NULLTRANSFORM) return (cmsHPROFILE) p;
@@ -1351,13 +1496,23 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
             // Yes, only named color. Create a named color-device
             // and append to named color table
 
-            p = cmsCreateTransform(hProfiles[0], dwInput, NULL, dwOutput, Intent, dwFlags);
+            cmsDeleteTransform((cmsHTRANSFORM) p);
+
+            p = (_LPcmsTRANSFORM) cmsCreateTransform(hProfiles[0], dwInput, NULL, dwOutput, Intent, dwFlags);
             for (i=1; i < nProfiles; i++) {
                     cmsReadICCnamedColorList(p, hProfiles[i], icSigNamedColor2Tag);
             }
 
             return p;   // Ok, done so far
     }
+    else
+    if (nNamedColor > 0) {
+
+        cmsDeleteTransform((cmsHTRANSFORM) p);
+        cmsSignalError(LCMS_ERRC_ABORTED, "Could not mix named color profiles with other types in multiprofile transform");
+        return NULL;
+    }
+
 
     // We will need a 3DCLUT for device link
     Grid =  cmsAllocLUT();
@@ -1424,7 +1579,7 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
                                                   Intent, dwPrecalcFlags);           
                 }
                 else {
-                        Transforms[i]  = cmsCreateTransform((ColorSpaceOut == icSigLabData ? hLab : hXYZ), FormatInput, 
+                        Transforms[i]  = cmsCreateTransform((ColorSpaceIn == icSigLabData ? hLab : hXYZ), FormatInput, 
                                                   hProfile, FormatOutput, 
                                                   Intent, dwPrecalcFlags);           
                     
@@ -1485,7 +1640,9 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
 
     // All ok, store the newly created LUT   
     p -> DeviceLink   = Grid;
-    p -> xform        = PrecalculatedXFORM;
+
+    SetPrecalculatedTransform(p, dwFlags);
+    
 
 
     for (i=nProfiles-1; i >= 0; --i)
